library walletpackage;

import 'dart:math';
import 'dart:typed_data';

import 'package:bitcoin_flutter/bitcoin_flutter.dart' as bitcoin_flutter;
import 'package:bip39/bip39.dart' as bip39;

import 'package:web3dart/crypto.dart' as web3dart_crypto;
import 'package:web3dart/web3dart.dart' as web3dart;

import 'package:web3dart/src/utils/rlp.dart' as rlp;
import 'package:web3dart/src/utils/typed_data.dart';

class WalletData {
  String mnemonic; //助记词，本地不保存
  String keystore;
  Map<String, String> addresses = new Map<String, String>();
}

class _CoinInfo {
  final String path;
  final bitcoin_flutter.NetworkType network;
  final int decimals;
  final String address;
  final int gasLimit;

  _CoinInfo(this.path, this.network, this.decimals,
      {this.address, this.gasLimit});
}

class Wallet {
  static final _litcoin = new bitcoin_flutter.NetworkType(
      messagePrefix: '\x19Litecoin Signed Message:\n',
      bech32: 'ltc',
      bip32: new bitcoin_flutter.Bip32Type(
          public: 0x019da462, private: 0x019d9cfe),
      pubKeyHash: 0x30,
      scriptHash: 0x32,
      wif: 0xb0);

  static final _bitcoincash = bitcoin_flutter.NetworkType(
      messagePrefix: '\x18Bitcoin Signed Message:\n',
      bech32: 'bc',
      bip32: new bitcoin_flutter.Bip32Type(
          public: 0x0488b21e, private: 0x0488ade4),
      pubKeyHash: 0x00,
      scriptHash: 0x05,
      wif: 0x80);

  static final Map<String, _CoinInfo> _bitcoinType = {
    'BTC': _CoinInfo("m/44'/0'/0'/0/0", bitcoin_flutter.bitcoin, 8),
    'LTC': _CoinInfo("m/44'/2'/0'/0/0", _litcoin, 8),
    'BCH': _CoinInfo("m/44'/145'/0'/0/0", _bitcoincash, 8),
  };
  static final Map<String, _CoinInfo> _omniType = {
    'USDT':
        _CoinInfo("m/44'/0'/0'/0/1", bitcoin_flutter.bitcoin, 8, address: '31'),
  };
  static final Map<String, _CoinInfo> _ethereumType = {
    'ETH': _CoinInfo("m/44'/60'/0'/0/0", null, 18, gasLimit: 21000),
    'ETC': _CoinInfo("m/44'/61'/0'/0/0", null, 18, gasLimit: 21000),
  };
  static final Map<String, _CoinInfo> _erc20Type = {
    'EUSDT': _CoinInfo("m/44'/60'/0'/0/1", null, 6,
        address: '0xdAC17F958D2ee523a2206206994597C13D831ec7', gasLimit: 60000),
  };
  static final Map<String, _CoinInfo> _allType = new Map<String, _CoinInfo>();

  static int chainId = null;
  static final innerSignature = chainId == null
      ? null
      : web3dart_crypto.MsgSignature(BigInt.zero, BigInt.zero, chainId);

  static bitcoin_flutter.HDWallet _getChildNode(
      bitcoin_flutter.HDWallet hdWallet, String type) {
    if (_allType.length == 0) {
      _allType
        ..addAll(_bitcoinType)
        ..addAll(_ethereumType)
        ..addAll(_omniType)
        ..addAll(_erc20Type);
    }

    if (_allType.containsKey(type)) {
      _CoinInfo coinInfo = _allType[type];
      print(coinInfo);
      hdWallet.network = coinInfo.network;
      bitcoin_flutter.HDWallet childnode = hdWallet.derivePath(coinInfo.path);
      return childnode;
    } else {
      print('_allType not contains key : $type');
      return null;
    }
  }

  static web3dart.EthereumAddress _node2EtherumAddress(
      bitcoin_flutter.HDWallet ethnode) {
    if (ethnode == null) return null;
    web3dart.EthereumAddress ethaddr = web3dart.EthereumAddress(web3dart_crypto
        .publicKeyToAddress(web3dart_crypto.privateKeyBytesToPublic(
            web3dart_crypto.hexToBytes(ethnode.privKey))));
    return ethaddr;
  }

  static void _addBitcoinAddress(
      WalletData wd, bitcoin_flutter.HDWallet hdWallet, String type) {
    bitcoin_flutter.HDWallet node = _getChildNode(hdWallet, type);
    wd.addresses.addAll({type: node.address});
  }

  static void _addEtherumAddress(
      WalletData wd, bitcoin_flutter.HDWallet hdWallet, String type) {
    bitcoin_flutter.HDWallet node = _getChildNode(hdWallet, type);
    wd.addresses.addAll({type: _node2EtherumAddress(node).hexEip55});
  }

  static Future<WalletData> createMnemonicAndAddress(String passwd) async {
    String mnemonic = bip39.generateMnemonic();
    WalletData walletData =
        await importMnemonicAndAddress(mnemonic, passwd: passwd);
    return walletData;
  }

  static Future<WalletData> importMnemonicAndAddress(String mnemonic,
      {String passwd = ''}) async {
    if (bip39.validateMnemonic(mnemonic)) {
      WalletData wd = new WalletData();
      wd.mnemonic = mnemonic;

      bitcoin_flutter.HDWallet hdWallet =
          bitcoin_flutter.HDWallet.fromSeed(bip39.mnemonicToSeed(wd.mnemonic));

      web3dart.Wallet wallet = web3dart.Wallet.createNew(
          web3dart.EthPrivateKey.fromHex(hdWallet.seed),
          passwd,
          Random.secure());
      wd.keystore = wallet.toJson();
//      print(wd.keystore);

      _bitcoinType.keys.forEach((type) {
        _addBitcoinAddress(wd, hdWallet, type);
      });
      _omniType.keys.forEach((type) {
        _addBitcoinAddress(wd, hdWallet, type);
      });
      _ethereumType.keys.forEach((type) {
        _addEtherumAddress(wd, hdWallet, type);
      });
      _erc20Type.keys.forEach((type) {
        _addEtherumAddress(wd, hdWallet, type);
      });
//      print(wd.addresses);

      return wd;
    } else {
      return null;
    }
  }

  static Future<String> changePassword(
      String keystore, String passwd, String newPasswd) async {
    try {
      web3dart.Wallet wallet = web3dart.Wallet.fromJson(keystore, passwd);
      web3dart.Wallet newWallet = web3dart.Wallet.createNew(
          wallet.privateKey, newPasswd, Random.secure());
      return newWallet.toJson();
    } catch (e) {
      print('signTransaction error : $e');
    }
    return null;
  }

  static Future<String> _signBitcoinTransaction(
      bitcoin_flutter.HDWallet wallet, String txhex) async {
    try {
      bitcoin_flutter.ECPair ecPair = bitcoin_flutter.ECPair.fromPrivateKey(
          web3dart_crypto.hexToBytes(wallet.privKey));
      bitcoin_flutter.Transaction transaction =
          bitcoin_flutter.Transaction.fromHex(txhex);
      bitcoin_flutter.TransactionBuilder transactionBuilder =
          bitcoin_flutter.TransactionBuilder.fromTransaction(transaction);
      transactionBuilder.inputs.forEach((element) {
        int index = transactionBuilder.inputs.indexOf(element);
        transactionBuilder.sign(vin: index, keyPair: ecPair);
      });
      String hex = transaction.toHex();
      return hex;
    } catch (e) {
      print('_signBitcoinTransaction error : $e');
    }
    return null;
  }

  static Future<String> _signEthereumTransaction(
      web3dart.EthPrivateKey privateKey, String txhex) async {
    Uint8List encoded = web3dart_crypto.hexToBytes(txhex);
    final web3dart_crypto.MsgSignature signature =
        await privateKey.signToSignature(encoded, chainId: chainId);
//    uint8ListFromList(rlp.encode(_encodeToRlp(transaction, signature)))
//    return '0x' + web3dart_crypto.bytesToHex(signature);
    return null;
  }

  static Future<String> signTransaction(
      String type, String keystore, String passwd, String txhex) async {
    try {
      web3dart.Wallet wallet = web3dart.Wallet.fromJson(keystore, passwd);
      bitcoin_flutter.HDWallet hdWallet =
          bitcoin_flutter.HDWallet.fromSeed(wallet.privateKey.privateKey);
      bitcoin_flutter.HDWallet childWallet = _getChildNode(hdWallet, type);
      String signedTxhex;
      if (childWallet != null) {
        if (_bitcoinType.keys.contains(type) || _omniType.keys.contains(type)) {
          print(childWallet.address);
          signedTxhex = await _signBitcoinTransaction(childWallet, txhex);
        } else if (_ethereumType.keys.contains(type) ||
            _erc20Type.keys.contains(type)) {
          print(_node2EtherumAddress(childWallet).hexEip55);
          web3dart.EthPrivateKey privateKey =
              web3dart.EthPrivateKey.fromHex(childWallet.privKey);
          signedTxhex = await _signEthereumTransaction(privateKey, txhex);
        } else {
          print('Unkown type : $type');
        }
      } else {
        print('childWallet is null');
      }
      return signedTxhex;
    } catch (e) {
      print('signTransaction error : $e');
    }
    return null;
  }

  static Future<String> createBitcoinTransaction(
      List ins, List outs, int fee) async {
    return null;
  }

  static Future<String> createEthereumTransaction(
      String from, String to, int value, int gasPrice) async {
    return null;
  }

  static Future<String> createOmniTransaction(
      List ins, List outs, int fee, int value, int contract) async {
    return null;
  }

  static List<dynamic> _encodeToRlp(web3dart.Transaction transaction,
      web3dart_crypto.MsgSignature signature) {
    final list = [
      transaction.nonce,
      transaction.gasPrice.getInWei,
      transaction.maxGas,
    ];

    if (transaction.to != null) {
      list.add(transaction.to.addressBytes);
    } else {
      list.add('');
    }

    list..add(transaction.value.getInWei)..add(transaction.data);

    if (signature != null) {
      list..add(signature.v)..add(signature.r)..add(signature.s);
    }

    return list;
  }

  static Future<String> createErc20Transaction(String type, String from,
      String to, int value, int gasPrice, int nonce) async {
    if (_erc20Type.keys.contains(type)) {
      _CoinInfo coinInfo = _erc20Type[type];

      String operationHex =
          '0xa9059cbb000000000000000000000000' + to.toLowerCase().substring(2);
      String valueHex = value.toRadixString(16);
      String tokenHex =
          ('0000000000000000000000000000000000000000000000000000000000000000' +
                  valueHex)
              .substring(valueHex.length);
      String dataHex = operationHex + tokenHex;

      web3dart.Transaction transaction = new web3dart.Transaction(
          from: web3dart.EthereumAddress.fromHex(from),
          to: web3dart.EthereumAddress.fromHex(coinInfo.address),
          maxGas: coinInfo.gasLimit,
          gasPrice:
              web3dart.EtherAmount.inWei(BigInt.from(gasPrice * 1000000000)),
          value: web3dart.EtherAmount.zero(),
          data: web3dart_crypto.hexToBytes(dataHex),
          nonce: nonce);

      final Uint8List encoded = uint8ListFromList(
          rlp.encode(_encodeToRlp(transaction, innerSignature)));
      return web3dart_crypto.bytesToHex(encoded);
    } else {
      return null;
    }
  }
}
